/* In this version, A-matrix is computed using doubles 
   only, but then stored using floats. When returning
   a single column of the matrix, the stored values
   are first converted to doubles and then returned. 
   If the program is used without the store option, 
   then everything is done as doubles. */

#include "A_comp_iupui.h"
#include <mat.h>
#include <matrix.h>

#undef LEN_Z
#define EVEN 0
#define ODD 1
#define NBALLOC 50

#define NORMALIZE_NRM
/*#define CONVOLVE_PIXEL*/

/* center of gantry rotation in cm */
double y_rot_ctr=0.0; /*0.07;*/ 
double x_rot_ctr=0.0; /* 0.0; */





static void A_comp_ij(struct sino_info *sino,struct geom_info *geom,struct column_flt **sp_col_arr, struct system_resp *sysres);
void read_glob_params(FILE *fp, struct system_resp *sysres, char *fname);

void read_resp_matrix(FILE *fp,struct resp_matrix *resmat,char *fname);
float **conv_pix(float **Horg,int N_t,int N_th,double step_t,struct geom_info *geom);


void init_A_matrix(struct sino_info *sino,struct geom_info *geom, char *fname_sysres)
{
  
  double inc_theta;
  double pi,rot_ang,rs,r_rot_ctr;
  int fr;
  int i,j;
  double x,y,x1,y1;
  struct system_resp sysres;
  
  /* Read system response file */
  read_system_response(fname_sysres,&sysres);
  
  if (sysres.analytical==1)
    printf("***** Using ANALYTICAL system response (specified in response file)\n\n");
  else if (sino->analyt_sysresp==1)
    {
      printf("***** Using ANALYTICAL system response (forced by cmd arg)\n\n");
      sysres.analytical=1;
    }
  else
    printf("Using MEASURED system response\n");
    

  calc_sino_arrays(sino,&sysres);
  printf("Initializing A-matrix...\n");
  fflush(stdout);fflush(stderr);

  /* Allocate sparse structure to store entire matrix */
  sino->sys_priv.sp_col_arr=(struct column_flt **)multialloc(sizeof(struct column_flt),2,geom->n_y,geom->n_x);


  /* This will only precompute system matrix for gantry pos. zero */
  A_comp_ij(sino,geom,sino->sys_priv.sp_col_arr,&sysres);
      

  /* precompute sin and cos for rotation angles 
     for different gantry positions */
  sino->sys_priv.cs_rot=(double*)mget_spc(sino->no_frames,sizeof(double));
  sino->sys_priv.si_rot=(double*)mget_spc(sino->no_frames,sizeof(double));

  pi = 4.0 * atan(1.0);
  inc_theta = pi / ((double)sino->no_frames);

  for (fr=0;fr<sino->no_frames;fr++)
    {
      rot_ang = inc_theta*(double)fr - sino->theta_0;

      sino->sys_priv.cs_rot[fr]=cos(rot_ang);
      sino->sys_priv.si_rot[fr]=sin(rot_ang);
      /*printf("rot_ang=%f\n",180*rot_ang/pi);*/
    }
      

  /* Compute valid radius for which we can compute reconstruction,
     considering center of gantry rotation */
  rs = sqrt(geom->field_of_view * geom->field_of_view / 4.0);
  r_rot_ctr = sqrt(x_rot_ctr*x_rot_ctr + y_rot_ctr*y_rot_ctr);
      
  rs -= r_rot_ctr;
  /*r -= geom->delta;*/
  sino->sys_priv.r_sq = rs*rs;
  

  printf("Radius of reconstruction is r=%f.\n",sqrt(sino->sys_priv.r_sq));

  /* calculate pixmask over which reconstruction occurs */
  sino->pix_reconst_mask=(short**)multialloc(sizeof(short),2,geom->n_y,geom->n_x);
  for (i=0;i<geom->n_y;i++)
    for (j=0;j<geom->n_x;j++) 
      {
	y = geom->y_0 + (double)i * geom->delta;
	x = geom->x_0 + (double)j * geom->delta;
    
	/*x-y coord. of pixel relative to center of gantry rotation*/ 
	x1 = x - x_rot_ctr;
	y1 = y - y_rot_ctr;
	if (x1*x1+y1*y1<sino->sys_priv.r_sq)
	  sino->pix_reconst_mask[i][j]=1;
	else
	  sino->pix_reconst_mask[i][j]=0;
    
      }
  
  fflush(stderr);
  printf("done initializing A_matrix.\n\n");fflush(stdout);
}


void free_A_matrix(struct sino_info *sino, struct geom_info *geom)
{
  int i,j;
  struct column_flt *pix_sp_col;

  for (i=0;i<geom->n_y;i++)
    for (j=0;j<geom->n_y;j++)
      {
	pix_sp_col = &sino->sys_priv.sp_col_arr[i][j];
	if (pix_sp_col->n_row > 0)
	  {
	    free((void*)pix_sp_col->val);
	    free((void*)pix_sp_col->row);
	  }
      }


  multifree((void*)sino->sys_priv.sp_col_arr,2);
  free((void*)sino->sys_priv.cs_rot);
  free((void*)sino->sys_priv.si_rot);

  free((void*)sino->t_arr);
  free((void*)sino->theta_arr);
  free((void*)sino->mask_arr);

  if (sino->nrm_FLAG)
    free((void*)sino->nrm_arr);

  if (sino->DECAY_FLAG)
    free((void*)sino->decay_arr);
  if (sino->SINO_COR_FLAG)
     free((void*)sino->sino_cor_arr);
}




/* Comments added 3/3/98 TF */
/*   Acomp computes one column of the A-matrix, i.e. takes pixel location row,col as*/
/*   input and returns sp_col , where */
/*   int *spcol.row   is a list of the projection indices */ 
/*(proj. in 1D order (angle*n_t + t)) that intersect with the pixel */
/*   double  sp_col.val[i]  contains the length of the intersection with projection spcol.row[i] */
/*  spcol.n_row is number of projections that intersect with pixel */
void A_comp(
	    int im_row, int im_col,
	    struct sino_info *sino,
	    struct geom_info *geom,
	    double **pix_prof,
	    struct column *sp_col, int res_level)
{


  static struct column_flt **sp_col_arr,*spc_f;
  static double *cs_rot,*si_rot;
  static double r_sq,x,y,x1,y1,x_rot,y_rot;
  static int r,nr,fr,row_rot,col_rot,rw;
  static int SINO_COR_FLAG, fr_offset;



  /* Copy sysmat data from struct into local variables
     for better readibility */
  sp_col_arr= sino->sys_priv.sp_col_arr;
  cs_rot    = sino->sys_priv.cs_rot;
  si_rot    = sino->sys_priv.si_rot;
  r_sq      = sino->sys_priv.r_sq;
  SINO_COR_FLAG=sino->SINO_COR_FLAG;
 

  /* res_level is not used by this version, arg is just for
     backward compatibility for calling format */
  if (res_level != 0)
    {
      fflush(stdout);
      fprintf(stderr,"A_comp_iupui not implemented for res_level != 0.\n");
      exit(-1);
    }


  /* Copy in prj's for ALL gantry positions */

  /* x-y location of pixel */
  y = geom->y_0 + (double)im_row * geom->delta;
  x = geom->x_0 + (double)im_col * geom->delta;

  /* x-y coord. of pixel relative to center of gantry rotation */
  x1 = x - x_rot_ctr;
  y1 = y - y_rot_ctr;
 

  sp_col->n_row = 0;
  sp_col->multiplier = 1.0;
  nr=0;
 

  if (x1*x1+y1*y1<r_sq)
    for (fr=0;fr<sino->no_frames;fr++)
      {
	/*printf("cs_rot(fr)=%f si_rot(fr)=%f x_rot_ctr=%f y_rot_ctr=%f \n",cs_rot[fr],si_rot[fr],x_rot_ctr,y_rot_ctr);*/
	x_rot = x1*cs_rot[fr] + y1*si_rot[fr] + x_rot_ctr;
	y_rot = -x1*si_rot[fr] + y1*cs_rot[fr] + y_rot_ctr;
	
	row_rot=(int)((y_rot-geom->y_0)/geom->delta + 0.5);
	col_rot=(int)((x_rot-geom->x_0)/geom->delta + 0.5);
	
	if ((row_rot<0) || (row_rot>=geom->n_y) || (col_rot<0) || (col_rot>=geom->n_x))
	  {
	    fprintf(stderr,"Error: row=%d row_rot=%d col=%d col_rot=%d\n",im_row,row_rot,im_col,col_rot);
	    printf("x=%f y=%f x_rot=%f y_rot=%f x1=%f y1=%f x1_rot=%f y1_rot=%f\n",x,y,x_rot,y_rot,x1,y1,x1*cs_rot[fr] + y1*si_rot[fr],-x1*si_rot[fr] + y1*cs_rot[fr]);
	    exit(-1);
	  }
	
	spc_f=&sp_col_arr[row_rot][col_rot];
	nr = spc_f->n_row;
	fr_offset = 6272*fr;

	if (SINO_COR_FLAG) 
	  {
	    for (r=0;r<nr;r++)
	      {
		rw = spc_f->row[r] + fr_offset;
		sp_col->row[r+sp_col->n_row] = rw;
		sp_col->val[r+sp_col->n_row] = sino->sino_cor_arr[rw] * (double)spc_f->val[r];
	      }
	  }
	else
	  {
	    for (r=0;r<nr;r++)
	      {
		sp_col->val[r+sp_col->n_row]=(double)spc_f->val[r];
		sp_col->row[r+sp_col->n_row]=spc_f->row[r] + fr_offset;
	      }
	  }
	sp_col->n_row += nr;
      }

  /*printf("sp_col->n_row=%d\n",sp_col->n_row);*/
}



static void A_comp_ij(struct sino_info *sino,struct geom_info *geom,struct column_flt **sp_col_arr, struct system_resp *sysres)
{
  int	ind,pr,pind,i;
  double Aval,ang,x,y,t_pix,t_i,diff_t,pi;
  int row,col;
  struct column_flt *pix_sp_col;
  int pix_proj_ct;
  int n_t,n_theta;

 
  double csa,ssa;
 
  double r_sq;

  /*#ifndef  ANALYTICAL*/
  int ind_t,ind_th;
  float **H;
  int N_th,N_t;
  double th_range,t_range,step_t,step_th;
  /*#endif*/


  r_sq=geom->field_of_view * geom->field_of_view / 4.0;
  printf("Using geom->delta=%f for approximation of integral.\n",geom->delta);


  /*if (fabs(sino->theta_0)>1e-30)
    {
    fprintf(stderr,"theta_0 must be 0\n");
    exit(-1);
    }
  */


  if (sysres->analytical==1)
    {
      printf("********* USING ANALYTICAL SYSTEM MATRIX **********\n");
      H=NULL;
      step_t=step_th=t_range=th_range=0.0;
      N_th=N_t=0;
    }
  else
    {
      N_th=sysres->resmat[0].N_th;
      N_t=sysres->resmat[0].N_t;
      th_range=sysres->resmat[0].th_range;
      t_range=sysres->resmat[0].t_range; /* in cm */
      
      step_t=2*t_range/(N_t-1);
      step_th=2*th_range/(N_th-1);

#ifdef CONVOLVE_PIXEL
      H=conv_pix(sysres->resmat[0].data,N_t,N_th,step_t,geom);
#else
      H=sysres->resmat[0].data;
#endif

      /*H=read_system_response("sysmat",&N_th,&N_t,&th_range,&t_range);
	t_range /= (double)10.0;*/
      /*printf("read system response file. N_th=%d, N_t=%d th_range=%f rad t_range=%f cm\n",N_th,N_t,th_range,t_range);*/
    }

  


 

  pi = 4.0 * atan(1.0);


  for (row=0;row<geom->n_y;row++)
    for (col=0;col<geom->n_x;col++)
      {
	sp_col_arr[row][col].multiplier = 1.0;
	sp_col_arr[row][col].n_row = 0;
      }
  
  /*printf("sino->prj_max_t=%f\n\n\n",sino->prj_max_t);*/
  /* first calc system matrix for frame 0 , i.e. gantry angle 0 degrees.*/
  n_t=112;
  n_theta=56;/* different from sino->n_theta */

  for(pr=0; pr < n_theta; pr++)
    for(i = 0; i < n_t; i++) /* i=0 is nonsense, same detectors */ 
      {   
	pind = pr*n_t;
	ind = pind+i; /* linear index in 1D array 
			 of all proj. (angle*n_t + t)*/	  
	t_i = sino->t_arr[ind]; /* parallel offset for this prj */

	if (sino->mask_arr[ind])
	  if (fabs(t_i) < sino->prj_max_t) /* don't use prj's outside of spec. radius */
	    {
	      ang = sino->theta_arr[ind]; /* angle of this prj */
	      
	      csa=cos(ang);
	      ssa=sin(ang);
	      
	      /*
		ind_rot=ind-fr*6272;
		t_prj_rot =sino->t_arr[ind_rot];
		th_prj_rot=sino->theta_arr[ind_rot];
		csr=cos(th_prj_rot);
		ssr=sin(th_prj_rot);
	      */
	      for (row=0;row<geom->n_y;row++)
		{
		  y = geom->y_0 + (double)row * geom->delta;
		  for (col=0;col<geom->n_x;col++)
		    {
		      /* center coord. of pixel */
		      x = geom->x_0 + (double)col*geom->delta;
		      
		      if (x*x+y*y < r_sq)
			{
			  t_pix=x*csa-y*ssa;
			  diff_t = t_i-t_pix;
			  
			  
			  /*
			    Now we are at the right pixel/projection comb., now 
			    will rotate pix-coordinates to gantry position zero and consider
			    projection of same detector pair from gantry position 0.
			  */
			  
			  /*x_rot = x*cs_rot + y*si_rot;
			    y_rot = -x*si_rot + y*cs_rot;
			    
			    
			    t_pix_rot=x_rot*csr - y_rot*ssr;
			    
			    m_t=t_prj_rot-t_pix_rot;
			  */
			  

			  if (sysres->analytical==1)
			    {
			      if (diff_t<0.0)
				diff_t = -diff_t;
			  
			      if (diff_t<sysres->ddet)
				{
				  Aval=0.1*(1.0-diff_t/sysres->ddet);
				}
			      else
				Aval=-1.0;
			    }
			  else
			    {
			      if ((diff_t>-t_range/2.0) && (diff_t<t_range/2.0))
				{
				  
				  ind_t = (int)( (diff_t+t_range)/step_t+0.5 );
				  ind_th= (int)( (ang+th_range)/step_th + 0.5);
				  
				  
				  if ((ind_t<0) || (ind_t>=N_t))
				    {
				      fprintf(stderr,"ind_t=%d, diff_t=%f, out of range.\n\a\n",ind_t,diff_t);
				      exit(-1);
				    }
				  if ((ind_th<0) || (ind_th>=N_th))
				    {
				      fprintf(stderr,"ind_th=%d, ang=%f, out of range.\n\a\n",ind_th,ang);
				      exit(-1);
				    }
				  
				  Aval=H[ind_th][ind_t];
				}
			      else
				Aval=-1.0;
			      
			    }
				
			  if (Aval > 0.0) 
			    {
			      /* get pointer to sp_col of this pixel */
			      pix_sp_col = &sp_col_arr[row][col];
			      pix_proj_ct = pix_sp_col->n_row;
			      
			      /* check if we have allocated enough non-zeros rows */
			      if (NBALLOC*(pix_proj_ct/NBALLOC)==pix_proj_ct)
				{
				  if (pix_proj_ct>0)
				    {
				      pix_sp_col->val = (float*)realloc((void*)pix_sp_col->val,(pix_proj_ct+NBALLOC)*sizeof(float));
				      pix_sp_col->row = (int*)realloc((void*)pix_sp_col->row,(pix_proj_ct+NBALLOC)*sizeof(int));
				    }
				  else
				    {
				      pix_sp_col->val = (float*)malloc((pix_proj_ct+NBALLOC)*sizeof(float));
				      pix_sp_col->row = (int*)malloc((pix_proj_ct+NBALLOC)*sizeof(int));
				    }
				  if ((pix_sp_col->val==NULL) || (pix_sp_col->row==NULL))
				    {
				      fflush(stdout);fflush(stderr);
				      fprintf(stderr,"Malloc error in A_comp.\n\a\n");fflush(stderr);
				      exit(-1);
				    }
				}
			      
			      
			      pix_sp_col->val[pix_proj_ct] = (float)Aval;
			      
			      
			      if (sino->nrm_FLAG)
				pix_sp_col->val[pix_proj_ct] *= sino->nrm_arr[ind];

			      
			      /* changed 11/15/99: for numerical integration
				 multiply by pix-area, so values of recon.
				 are consistent for multiscale case 50*  */
			      pix_sp_col->val[pix_proj_ct] *= (160.0*geom->delta*geom->delta);
			      
			      
			      pix_sp_col->row[pix_proj_ct] = ind;/* Index for ray    */
			      pix_sp_col->n_row = pix_sp_col->n_row + 1;
			      
			    
			    } /* end if Aval > 0.0 */
			      
			} /* if x*x+y*y<r_sq */
			  
		    } /* end for col */
		} /* end for row */
	    } /* end if (mask_arr[ind]) */
      } /* end for proj. i */

  if (sysres->analytical!=1)
    multifree((void*)H,2);

#ifdef CONVOLVE_PIXEL
  multifree((void*)sysres->resmat[0].data,2);
#endif

}/* Close A_comp()	*/
  





/* Routine below uses IUPUI coordinate system: 
   pos. x points right
   pos. y points up
   theta: 0 for proj. parallel to y-axis, >0 CCW from y-axis, <- CW from y-axis
   then: t=x*cos(theta) + y*sin(theta); 
   NOTICE: in A_comp ABOVE, we use y pointing down, so need to use 
           t=x*cos(theta) - y*sin(theta) there.
*/

void calc_sino_arrays(struct sino_info *sino,struct system_resp *sysres)
{
  int det1[112][56], det2[112][56];
  int i,j,nprj,fr,dtop,dbot,ind;
  double ddet,rdet,theta,det_depth;
  double xc[112],yc[112];
  double inc_theta;
  int no_frames;
  double sep_d,mn=100.0,mx=-100.0,pi;
  double x_top,x_bot,y_top,y_bot;
  double cs_rot, si_rot, theta_offset;
  float min_nrm,max_nrm;
  float mean_nrm,ct;
  double tau,t_mid,corr_fact;
 
  no_frames=sino->no_frames;

  if (fabs(sino->sep_d-sysres->sep_d)>0.1)
    {
      fprintf(stderr,"WARNING: detector separation in sinogram %f different from that in system response file of %f.\nUsing the one in the response file\n\a\n",sino->sep_d,sysres->sep_d);
    }
  
  ddet=sysres->ddet;
  rdet=sysres->rdet;
  det_depth = sysres->dthck;
  sino->dthck=det_depth;
  sep_d=sysres->sep_d;
  printf("Parameters from system response file: \ndetector witdh ddet=%f\nradius of det-surface rdet=%f\ndet_depth=%f\nseparation sep_d=%f\n",ddet, rdet,det_depth,sep_d);
  rdet += 0.5*det_depth;
  sep_d += det_depth;
  
  printf("With added detector depth, use effective parameters:\nrdet=%f\nsep_d=%f\n",rdet,sep_d);


  printf("no_frames=%d\n",no_frames);
  
  pi = 4.0 * atan(1.0);

  sino->n_t = 112; /* 112 par. proj */
  sino->n_theta = no_frames*56; /* 56 angles per gantry position */

  /* increment of angle with each gantry position is 180deg/no_frames*/
  inc_theta = pi / ((double)no_frames); /* pi / ((double)no_frames-1.0);*/

  nprj=sino->n_t*sino->n_theta;

  sino->t_arr = (double*)get_spc(nprj,sizeof(double));
  sino->theta_arr = (double*)get_spc(nprj,sizeof(double));
  sino->mask_arr = (int*)get_spc(nprj,sizeof(int));
  
  if (sino->nrm_FLAG)
    sino->nrm_arr = (float*)get_spc(nprj,sizeof(float));
  else
     sino->nrm_arr = NULL;
  
  if (sino->DECAY_FLAG)
    sino->decay_arr = (float*)get_spc(nprj,sizeof(float));
  else
    sino->decay_arr = NULL;

  if (sino->attn.ATTN_MODE)
    sino->attn.attn_arr = (double*)get_spc(nprj,sizeof(double));
  else
    sino->attn.attn_arr = NULL;

  if (sino->DEADTIME_FLAG)
    sino->dtimecr_arr = (double*)get_spc(nprj,sizeof(double));
  else
    sino->dtimecr_arr = NULL;
  

  /* Compute detector pairs corresponding to sinogram entries */
  for (j=0;j<56;j++)
    for (i=0;i<112;i++)
      {
	det1[i][j] = ( (int)(84.0+(double)j+((double)i)/2.0) )%112; /* dbot*/
	det2[i][j] = ( (int)(84.0+(double)j-((double)i)/2.0) )%112; /* dtop*/
      }
 

  /* compute x-y coordinates for each detector */
  /* Using standard coordinate system: x points right, y points up */
  for (i=0;i<56;i++)
    {
      theta = ( (double)(i-28) + 0.5 )*ddet/rdet;
      xc[i] = -rdet*sin(theta);
      yc[i]= rdet * (1.0-cos(theta)) - sep_d/2.0;
      xc[i+56] = -xc[i];
      yc[i+56] = -yc[i];
    }
  /*printf("xc[0]=%f yc[0]=%f xc[97]=%f yc[97]=%f xc[63]=%f yc[63]=%f \n",xc[0],yc[0],xc[97],yc[97],xc[63],yc[63]);*/



  for (fr=0;fr<no_frames;fr++)
    {
      theta_offset=-1.0*inc_theta*(double)fr;
      cs_rot=cos(theta_offset);
      si_rot=sin(theta_offset);

      for (j=0;j<56;j++)
	for (i=0;i<112;i++)
	  {
	    dtop=det2[i][j];
	    dbot=det1[i][j];
	    ind=112*j+i+fr*6272;
	    if (dtop !=dbot)
	      {
		/* theta=0 is y-axis. CCW from y-axis have theta>0,
		   CW from y-axis have theta<0,
		   want theta to be in range -pi/2 <= theta < pi/2
		*/

		/* Apply rotation matrix to obtain x-y coordinates
		   for top and bottom detector for projection (j,i)
		   for this gantry position. */
		x_top = xc[dtop]*cs_rot - yc[dtop]*si_rot;
		y_top = xc[dtop]*si_rot + yc[dtop]*cs_rot;
		x_bot = xc[dbot]*cs_rot - yc[dbot]*si_rot;
		y_bot = xc[dbot]*si_rot + yc[dbot]*cs_rot;

		if (fabs(y_top-y_bot)>1e-30)
		  {
		    sino->theta_arr[ind] = atan( -(x_top-x_bot)/(y_top-y_bot) ); 
		    sino->t_arr[ind] = x_top*cos(sino->theta_arr[ind]) + y_top*sin(sino->theta_arr[ind]);
		  }
		else
		  {
		    sino->theta_arr[ind] = - pi/2.0;
		    sino->t_arr[ind] = x_top*cos(sino->theta_arr[ind]) + y_top*sin(sino->theta_arr[ind]);
		  }
		if (sino->theta_arr[ind]<mn)
		  mn=sino->theta_arr[ind];
		if (sino->theta_arr[ind]>mx)
		  mx=sino->theta_arr[ind];
		
		/* t-offset for projection. To the right from origin (+x)
		   is t>0, to the left is t<0
		*/
		
	      }
	    else if (i!=0) /* top and bottom detector are the same..no prj really*/
	      {
		printf("Error: dtop=dbot=%d and i= %d !=0.\n\a\n",dtop,i);
		exit(-1);
	      }
	    else
	      {
		/* The way iupui set up their sinograms is that 
		   for i=0, both detectors are the same */
		/* never use these prj's, they have single detector ! */
		sino->t_arr[ind]=1000.0;
		sino->theta_arr[ind]=0.0;
	      }
	  }
    }
  printf("sino->theta_0=%f  mn=%f mx=%f\n",sino->theta_0,mn,mx);
  
  
  /*
    j=1;
    for (fr=0;fr<2;fr++)
    for (i=0;i<112;i++) 
    {
    ind=112*j+i+fr*6272;
    printf("fr=%d i=%d j=%d sino->t_arr[ind]=%f\n",fr,i,j,sino->t_arr[ind]);
    }
  */

  if (sino->nrm_FLAG)
    {
      printf("Calculating normalization for all frames.\n");
      /* copy nrm into all arrays */
      for (j=0;j<56;j++)
	for (i=0;i<112;i++)
	  {	
	    ind=112*j+i;
	    sino->nrm_arr[ind]=(float)(sino->nrm_img[j][i]* cos(sino->theta_arr[ind]));
	  }
    }
  

  /* Calculate deadtime-correction */
  if (sino->DEADTIME_FLAG)
    calculate_deadtime_correction(sino, no_frames);



      
  /* compute mask */
  for (j=0;j<56;j++)
    for (i=0;i<112;i++)
      {
	ind=112*j+i;
	if ((det1[i][j]<=55)&&(det2[i][j]>=56))
	  sino->mask_arr[ind]=1;
	else
	  sino->mask_arr[ind]=0;
      }
  


  /* copy same mask to other frames */
  for (fr=1;fr<no_frames;fr++)
    for (i=0;i<6272;i++)
      {
	ind=fr*6272+i;
	sino->mask_arr[ind]=sino->mask_arr[i];
      }


  


 

  /* FIND min and mean of normalization sinogram*cos() after mask */
  ct=0.0;
  if (sino->nrm_FLAG)
    {
      min_nrm=1e3;
      max_nrm=-1e3;
      mean_nrm=0.0;
      
      for (j=0;j<56;j++)
	for (i=0;i<112;i++)
	  {	
	    ind=112*j+i;
	    if (sino->mask_arr[ind])
	      {
		if (min_nrm>sino->nrm_arr[ind])
		  min_nrm=sino->nrm_arr[ind];
		if (max_nrm<sino->nrm_arr[ind])
		  max_nrm=sino->nrm_arr[ind];
	    
		mean_nrm+=sino->nrm_arr[ind];
		ct++;
	      }
	  }
      
      mean_nrm /= ct; 
      printf("Minimum Normalization const = %f, mean=%f, max=%f\n",min_nrm,mean_nrm,max_nrm);
#ifdef NORMALIZE_NRM
      printf("normalize_nrm: divide nrm_array by %f\n",mean_nrm);
      for (j=0;j<56;j++)
	for (i=0;i<112;i++)
	  {	
	    ind=112*j+i;
	    sino->nrm_arr[ind] /= mean_nrm;
	  }
#endif
      
      
      /* copy same nrm to other frames */
      for (fr=1;fr<no_frames;fr++)
	for (i=0;i<6272;i++)
	  {
	    ind=fr*6272+i;
	    sino->nrm_arr[ind]=sino->nrm_arr[i];
	  }
    } 
  
  /* Now do correction for exp. decay */
  if (sino->DECAY_FLAG)
    {
      printf("Calculating decay correction...\n");fflush(stdout);
      tau=sino->halflife/log(2.0);
      printf("tau=%f\nframe_correction=[",tau);
      
      for (fr=0;fr<sino->no_frames;fr++)
	{
	  t_mid = sino->frames_starttime[0][fr]+0.5*sino->frames_duration[0][fr];
	  corr_fact = exp(-t_mid/tau);
	  printf("%f ",corr_fact);
	  for (i=0;i<6272;i++)
	    {	
	      ind=i+fr*6272;
	      sino->decay_arr[ind] = (float)corr_fact;
	    }
	}
      printf("]\n\n");
    }

  if (sino->attn.ATTN_MODE)
    calculate_attenuation_correction(sino);




  /* Consolidate ATTN and deadtime and DECAY correction into one
     array for faster multiplication in A_comp */
  sino->SINO_COR_FLAG=0;
  if ((sino->attn.ATTN_MODE) || (sino->DEADTIME_FLAG) || (sino->DECAY_FLAG))
    {
      sino->SINO_COR_FLAG=1;
      sino->sino_cor_arr=(double*)mget_spc(nprj,sizeof(double));

      for (i=0;i<nprj;i++)
	sino->sino_cor_arr[i] = 1.0;

      if (sino->attn.ATTN_MODE)
	for (i=0;i<nprj;i++)
	  sino->sino_cor_arr[i] *= sino->attn.attn_arr[i] ;
      
      if (sino->DEADTIME_FLAG)
	for (i=0;i<nprj;i++)
	  sino->sino_cor_arr[i] *= sino->dtimecr_arr[i];
      
      if (sino->DECAY_FLAG)
	for (i=0;i<nprj;i++)
	  {
	    sino->sino_cor_arr[i] *= sino->decay_arr[i];
	    /*printf("%d %f\n",i/6272, sino->decay_arr[i]);*/
	  }
    }


  
  
      /*  sino->SINO_COR_FLAG=0;
	  if ((sino->attn.ATTN_MODE) || (sino->DEADTIME_FLAG))
	  {
	  sino->SINO_COR_FLAG=1;
	  if ((sino->attn.ATTN_MODE) && (sino->DEADTIME_FLAG))
	  {
	  sino->sino_cor_arr=(double*)mget_spc(nprj,sizeof(double));
	  for (i=0;i<nprj;i++)
	  sino->sino_cor_arr[i] = sino->dtimecr_arr[i]*sino->attn.attn_arr[i];
	  free((void*)sino->dtimecr_arr);
	  free((void*)sino->attn.attn_arr);
	  }
	  else if (sino->attn.ATTN_MODE)
	  sino->sino_cor_arr = sino->attn.attn_arr;
	  else if (sino->DEADTIME_FLAG)
	  sino->sino_cor_arr = sino->dtimecr_arr;
	  else
	  {fprintf(stderr,"Internal error in calc_sino_arr.\n\a\n");exit(-1);}
	  
	  }
      */
    
}







void read_system_response(char *fname, struct system_resp *sysres)
{
  FILE *fp;
  char st1[256],st2[256];
  double version;
  int stat,m;


  printf("Reading system response file %s.\n",fname);

  fp=fopen(fname,"r");
  if (fp==NULL)
    {
      fprintf(stderr,"Can't open/read system_response file: %s\n\a\n",fname);
      exit(-1);
    }
  
  /* read header */
  stat=fscanf(fp,"%s %s %lf",st1,st2,&version);checkstat(stat,3);
  if ((strcmp(st1,"SystemResponseFile")) || (strcmp(st2,"Version")))
    {
      fprintf(stderr,"System response file %s doesn't start w/ correct header.\n\a\n",fname);
      exit(-1);
    }
  
  /* See if 'GeneralInfo' section is next, but skip over contents of section */
  stat=fscanf(fp,"%s",st1);checkstat(stat,1);
  if (strcmp(st1,"BeginGeneralInfo"))
    {
      fprintf(stderr,"System response file %s doesn't start w/ correct header: Second line doesn't have 'BeginGeneralInfo'\n\a\n",fname);
      exit(-1);
    }
  
  while (strcmp(st1,"EndGeneralInfo"))
    stat=fscanf(fp,"%s",st1);checkstat(stat,1);
  
  /* read section 'GlobalParameters' */
  read_glob_params(fp, sysres,fname);
  printf("Number of system response matrices = %d\n",sysres->no_matrices);
  
  if (sysres->no_matrices>0)
    {
      /* allocate matrices in structure */
      sysres->resmat = (struct resp_matrix*)get_spc(sysres->no_matrices,sizeof(struct resp_matrix));
  
  
      /* read matrices */
      for (m=0;m<sysres->no_matrices;m++)
	{
	  printf("Reading matrix %d.\n",m);
	  read_resp_matrix(fp,&sysres->resmat[m],fname);
	}
    }

  /* see if file ended */
  stat=fscanf(fp,"%s",st1);checkstat(stat,1);

  if (strcmp(st1,"EndSystemResponseFile"))
    {
      fprintf(stderr,"Error reading system response file: Don't read 'EndSystemResponseFile' where it should be. Instead %s.\n\a\n\n",st1);
      exit(-1);
    }
  
  fclose(fp);
  
}


void checkstat(int stat,int num)
{
  if (stat!=num)
    {
      fprintf(stderr,"Read error in system response file.\n\a\n");
      exit(-1);
    }
}



void read_glob_params(FILE *fp, struct system_resp *sysres,char *fname)
{
  int got_no_matrices=0,got_ddet=0;
  int got_dthck=0,got_rdet=0,got_sep_d=0,got_analyt=0;
  int stat;
  char st1[256];
  
  stat=fscanf(fp,"%s",st1);checkstat(stat,1);
  if (strcmp(st1,"BeginGlobalParameters"))
    {
      fprintf(stderr,"System response file %s doesn't start w/ correct header: Second line doesn't have 'BeginGlobalParameters'\n\a\n",fname);
      exit(-1);
    }
	
  for (;;)
    {
      stat=fscanf(fp,"%s",st1);checkstat(stat,1);
      if (!strcmp(st1,"EndGlobalParameters"))
	break;
      else if (!strcmp(st1,"use_analytical_response:"))
	{
	  stat=fscanf(fp,"%d",&sysres->analytical);checkstat(stat,1);
	  got_analyt=1;
	}      
      else if (!strcmp(st1,"number_response_matrices:"))
	{
	  stat=fscanf(fp,"%d",&sysres->no_matrices);checkstat(stat,1);
	  got_no_matrices=1;
	}      
      else if (!strcmp(st1,"detector_width_cm:"))
	{
	  stat=fscanf(fp,"%lf",&sysres->ddet);checkstat(stat,1);
	  got_ddet=1;
	}
      else if (!strcmp(st1,"detector_thickness_cm:"))
	{
	  stat=fscanf(fp,"%lf",&sysres->dthck);checkstat(stat,1);
	  got_dthck=1;
	}
      else if (!strcmp(st1,"radius_detector_curvature_cm:"))
	{
	  stat=fscanf(fp,"%lf",&sysres->rdet);checkstat(stat,1);
	  got_rdet=1;
	}
      else if (!strcmp(st1,"detector_separation_cm:"))
	{
	  stat=fscanf(fp,"%lf",&sysres->sep_d);checkstat(stat,1);
	  got_sep_d=1;
	}
      else
	{
	  fprintf(stderr,"Read error system response: Unrecognized global parameter %s\n\a\n",st1);
	  exit(-1);
	}
    }

  if (got_analyt!=1)
    {
      fprintf(stderr,"Error in file %s: use_analytical_response  not specified.\n\a\n",fname);
      exit(-1);
    } 
  
 if (got_no_matrices!=1)
    {
      fprintf(stderr,"Error in file %s: number_response_matrices not specified.\n\a\n",fname);
      exit(-1);
    } 

  
  if (got_ddet!=1)
    {
      fprintf(stderr,"Error in file %s: detector_width_cm not specified.\n\a\n",fname);
      exit(-1);
    }

  if (got_dthck!=1)
    {
      fprintf(stderr,"Error in file %s: detector_thickness_cm not specified.\n\a\n",fname);
      exit(-1);
    }

  if (got_rdet!=1)
    {
      fprintf(stderr,"Error in file %s: radius_detector_curvature_cm not specified.\n\a\n",fname);
      exit(-1);
    }
  
  if (got_sep_d!=1)
    {
      fprintf(stderr,"Error in file %s: detector_separation_cm: not specified.\n\a\n",fname);
      exit(-1);
    }

  if ((sysres->analytical==1) && (sysres->no_matrices!=0))
    {
      fprintf(stderr,"Error reading system response file: Analytical response specified, but number of matrices not zero but %d.\n\a\n",sysres->no_matrices);
      exit(-1);
    }
  
  if ((sysres->analytical==0) && (sysres->no_matrices<=0))
    {
      fprintf(stderr,"Error reading system response file: Analytical response NOT specified, but number of matrices zero.\n\a\n");
      exit(-1);
    }
}
    

void read_resp_matrix(FILE *fp,struct resp_matrix *resmat,char *fname)
{
  int stat;
  char st1[256];
  int got_x_pos=0,got_y_pos=0,got_N_th=0,got_N_t=0;
  int got_th_range=0,got_t_range=0;
  int i,j;

  stat=fscanf(fp,"%s",st1);checkstat(stat,1);
  if (strcmp(st1,"BeginMatrix"))
    {
      fprintf(stderr,"Can't read matrix: System response file %s doesn't have 'BeginMatrix'.\n\a\n",fname);
      exit(-1);
    }
  
  for (;;)
    {
      stat=fscanf(fp,"%s",st1);checkstat(stat,1);
      if (!strcmp(st1,"BeginMatrixData"))
	break;
      else if (!strcmp(st1,"x_pos_calibrated:"))
	{
	  stat=fscanf(fp,"%lf",&resmat->x_pos);checkstat(stat,1);
	  got_x_pos=1;
	}      
      else if (!strcmp(st1,"y_pos_calibrated:"))
	{
	  stat=fscanf(fp,"%lf",&resmat->y_pos);checkstat(stat,1);
	  got_y_pos=1;
	}      
      else if (!strcmp(st1,"size_N_theta:"))
	{
	  stat=fscanf(fp,"%d",&resmat->N_th);checkstat(stat,1);
	  got_N_th=1;
	}      
      else if (!strcmp(st1,"size_N_t:"))
	{
	  stat=fscanf(fp,"%d",&resmat->N_t);checkstat(stat,1);
	  got_N_t=1;
	}      
      else if (!strcmp(st1,"theta_range:"))
	{
	  stat=fscanf(fp,"%lf",&resmat->th_range);checkstat(stat,1);
	  got_th_range=1;
	}      
      else if (!strcmp(st1,"t_range:"))
	{
	  stat=fscanf(fp,"%lf",&resmat->t_range);checkstat(stat,1);
	  got_t_range=1;
	}      
      else
	{
	  fprintf(stderr,"Error reading system response file: Unrecognized matrix parameter %s\n\a\n",st1);
	  exit(-1);
	}
    }

  if (got_x_pos!=1)
    {
      fprintf(stderr,"Read Error system response: Matrix doesn't specify x_pos_calibrated.\n\a\n");
      exit(-1);
    }
  if (got_y_pos!=1)
    {
      fprintf(stderr,"Read Error system response: Matrix doesn't specify y_pos_calibrated.\n\a\n");
      exit(-1);
    }
  if (got_N_th!=1)
    {
      fprintf(stderr,"Read Error system response: Matrix doesn't specify size_N_theta\n\a\n");
      exit(-1);
    }
  if (got_N_t!=1)
    {
      fprintf(stderr,"Read Error system response: Matrix doesn't specify size_N_t\n\a\n");
      exit(-1);
    } 
  if (got_th_range!=1)
    {
      fprintf(stderr,"Read Error system response: Matrix doesn't specify theta_range\n\a\n");
      exit(-1);
    }
  if (got_t_range!=1)
    {
      fprintf(stderr,"Read Error system response: Matrix doesn't specify t_range\n\a\n");
      exit(-1);
    }
  

  /* allocate data matrix */
  resmat->data=(float**)multialloc(sizeof(float),2,resmat->N_th,resmat->N_t);

  /* read data array */
  for (i=0;i<resmat->N_th;i++)
    for (j=0;j<resmat->N_t;j++)
      {
	stat=fscanf(fp,"%f",&resmat->data[i][j]);
	checkstat(stat,1);
      }


  stat=fscanf(fp,"%s",st1);checkstat(stat,1);
  if (strcmp(st1,"EndMatrixData"))
    {
      fprintf(stderr,"Error reading system response file: 'EndMatrixData' statement not found. Instead %s. Maybe wrong size data array?\n\a\n",st1);
      exit(-1);
    }

  stat=fscanf(fp,"%s",st1);checkstat(stat,1);
  if (strcmp(st1,"EndMatrix"))
    {
      fprintf(stderr,"Error reading system response file: 'EndMatrix' statement not found. Instead %s.\n\a\n",st1);
      exit(-1);
    }

  


}



void calculate_attenuation_correction(struct sino_info *sino)
{
  double r,c,rsq;
  int i,j,fr,no_frames;
  double img_rows_dbl,angle,offset,row_dbl;
  double col_dbl, wr1,wr2,wc1,wc2;
  int row1,row2,ind,col1,col2;
  double pi;
  double attn_cor;
  double *theta_arr,*t_arr;

  pi = 4.0 * atan(1.0);
  no_frames=sino->no_frames;
  theta_arr=sino->theta_arr;
  t_arr=sino->t_arr;


  if (sino->attn.ATTN_MODE==2) /* analytical circle */
    {
      printf("Calculating analytical attenuation correction for circle of radius %fcm...",sino->attn.attn_analyt_r);fflush(stdout);
      r=sino->attn.attn_analyt_r;
      c=sino->attn.attn_cnst;
      rsq = r*r;
      for (i=0;i<6272;i++)
	if ((sino->t_arr[i]>-r) && (sino->t_arr[i]<r))
	  sino->attn.attn_arr[i] = exp(-c*2.0*sqrt(rsq-sino->t_arr[i]*sino->t_arr[i]));
	else
	  sino->attn.attn_arr[i] = 1.0;
    
      for (fr=1;fr<no_frames;fr++)
	for (i=0;i<6272;i++)
	  {
	    ind=fr*6272+i;
	    sino->attn.attn_arr[ind]=sino->attn.attn_arr[i];
	  }
      
      printf("done.\n");fflush(stdout);fflush(stderr);
    }
  else if (sino->attn.ATTN_MODE==1) /* sin - file */
    {
      img_rows_dbl = (double)sino->attn.img_rows;
      if (sino->attn.img_cols != 112)
	{
	  fflush(stdout);fflush(stderr);
	  fprintf(stderr,"Error in A_comp_iupui.c: attn correction only implemented for sin-atn files with 112 columns\n\a\n");
	  exit(-1);
	}

      for (fr=0;fr<no_frames;fr++)
	for (j=0;j<56;j++)
	  for (i=0;i<112;i++)
	    if (sino->mask_arr[112*j+i])
	      {
		ind=112*j+i+fr*6272;
				
		angle=theta_arr[ind];
		offset=t_arr[ind];
		row_dbl=(pi/2-angle)*img_rows_dbl/pi;
		col_dbl=56.0 + offset/0.165;
		row1=(int)row_dbl;
		row2=(int)(row_dbl+1.0);
		col1=(int)col_dbl;
		col2=(int)(col_dbl+1.0);
		wr2= row_dbl-(double)row1;
		wr1= 1.0-wr2;
		wc2= col_dbl-(double)col1;
		wc1= 1.0-wc2;
		
		
		if ((col1<=0) || (col2>=112))
		  {
		    fprintf(stderr,"Error calc attn_cor: col out of range, i=%d, j=%d, fr=%d, offset=%f, col_dbl=%f,  col1=%d col2=%d\n\a\n",i,j,fr,offset,col_dbl,col1,col2);
		    exit(-1);
		  }
		
		if (fabs(wr1*wc1+wr1*wc2+wr2*wc1+wr2*wc2-1.0)>1e-9)
		  {
		    fprintf(stderr,"Error calc. weights, sum=%f\n",wr1+wr2+wc1+wc2);
		    exit(-1);
		  }
		
		if (row2==img_rows_dbl)
		  {
		    attn_cor = wc1*sino->attn.sin_img[row1][col1]+wc2*sino->attn.sin_img[row1][col2]; 
		  }
		else
		  {
		    attn_cor = wr1*wc1*sino->attn.sin_img[row1][col1] + wr1*wc2*sino->attn.sin_img[row1][col2] + wr2*wc1*sino->attn.sin_img[row2][col1] + wr2*wc2*sino->attn.sin_img[row2][col2];
		    
		  }
		sino->attn.attn_arr[ind] = 1.0/attn_cor;
	      }
	    else
	      sino->attn.attn_arr[112*j+i+fr*6272] = 1.0;
      
    }
  else
    {
      fprintf(stderr,"Error: Invalid sino->attn.ATTN_MODE=%d.\n\a\n",sino->attn.ATTN_MODE);
      exit(-1);
    }

  
}




void calculate_deadtime_correction(struct sino_info *sino,int no_frames)
{
  int **frames_midtime_cor;
  double tau,w,num,den;
  int blk,fr,file,i,j,ind;
  double **dt_cor;
  int **blk1,**blk2; /* arrays to store the two detector 
			blocks 1..4 for each projection */


  dt_cor=(double**)multialloc(sizeof(double),2,4,no_frames);
  blk1=(int**)multialloc(sizeof(int),2,112,56);
  blk2=(int**)multialloc(sizeof(int),2,112,56);


  /* Calc. weighted average of deadtime correction over the
     sinogram input files. The average is calculated separately
     for each frame. Weights are the frame_duration and decay for
     the particular frame in each input file */
  frames_midtime_cor=calculate_frame_times(sino,no_frames);
  tau=sino->halflife/log(2.0);
  for (blk=0;blk<4;blk++)
    for (fr=0;fr<no_frames;fr++)
      {
	num=den=0.0;
	for (file=0;file<sino->dtime.no_sino_files;file++)
	  {
	    w=sino->frames_duration[file][fr] * exp(-frames_midtime_cor[file][fr]/tau);
	    num = num + w * sino->dtime.uncor_singles[file][fr][blk] / sino->dtime.cor_singles[file][fr][blk];
	    den += w;
	    /*printf("file=%d fr=%d w=%f rat=%f\n",file,fr,w,sino->dtime.uncor_singles[file][fr][blk] / sino->dtime.cor_singles[file][fr][blk]);*/
	  }
	dt_cor[blk][fr]=num/den;
	/*printf("**** blk=%d fr=%d dt_cor=%f\n",blk,fr,dt_cor[blk][fr]);*/
      }
  
  calc_deadtime_blocks(blk1,blk2);
  
  
  for (fr=0;fr<no_frames;fr++)
    for (j=0;j<56;j++)
      for (i=0;i<112;i++)
	{
	  ind=112*j+i+fr*6272;
	  sino->dtimecr_arr[ind] = dt_cor[blk1[i][j]][fr] *  dt_cor[blk2[i][j]][fr];
	}





  multifree((void*)frames_midtime_cor,2);
  multifree((void*)dt_cor,2);
  multifree((void*)blk1,2);
  multifree((void*)blk2,2);

}
  
void calc_deadtime_blocks(int **det1, int **det2)
{
  int i,j;

  for (j=0;j<56;j++)
    for (i=0;i<112;i++)
      {
	/* calculate detector pair for sinogram entry i,j */
	det1[i][j] = ( (int)(84.0+(double)j+((double)i)/2.0) )%112; /* dbot*/
	det2[i][j] = ( (int)(84.0+(double)j-((double)i)/2.0) )%112; /* dtop*/
      
	/* take (int)(det/28) to calculate which detector block
	   det1 and det2 are in. Blocks are 0-3, starting w/ 0 lower
	   right and then clockwise around scanner */
	det1[i][j] = det1[i][j]/28;
	det2[i][j] = det2[i][j]/28;
      }
}


int** calculate_frame_times(struct sino_info *sino, int no_frames)
{
  int file, fr;
  int **frames_midtime_cor;
  int old, starttime;
  
  frames_midtime_cor=(int**)multialloc(sizeof(int),2,sino->dtime.no_sino_files,no_frames);
  
  /* check if month changed */
  for (file=1;file<sino->dtime.no_sino_files;file++)
    if (sino->dtime.scan_start_day[file] < sino->dtime.scan_start_day[file-1])
      {
	/* if month really changed, day should go from at least 28 down to 1 */
	if (sino->dtime.scan_start_day[file] - sino->dtime.scan_start_day[file-1] <= -27)
	  {
	    /* month changed */
	    sino->dtime.scan_start_day[file] = sino->dtime.scan_start_day[file-1];
	    sino->dtime.scan_start_hour += 24;
	  }
	else
	  {
	    /* filenames not in order of scans taken */
	    fprintf(stderr,"Error: scan_start_day[file %d]=%d < %d=scan_start_day[file %d]\n",file,sino->dtime.scan_start_day[file],sino->dtime.scan_start_day[file-1],file-1);
	    exit(-1);
	  }
      }
  
  
  for (file=0;file<sino->dtime.no_sino_files;file++)
    for (fr=0;fr<no_frames;fr++)
      frames_midtime_cor[file][fr] =  86400*sino->dtime.scan_start_day[file] + 3600*sino->dtime.scan_start_hour[file] + 60*sino->dtime.scan_start_minute[file] + sino->dtime.scan_start_second[file] + sino->frames_starttime[file][fr] + 0.5*sino->frames_duration[file][fr];
  
  old=0;
  starttime = frames_midtime_cor[0][0];
  for (file=0;file<sino->dtime.no_sino_files;file++)
    for (fr=0;fr<no_frames;fr++)
      {
	frames_midtime_cor[file][fr] -= starttime;
	if (frames_midtime_cor[file][fr]<old)
	  {
	    fprintf(stderr,"Error: scan time of frame %d in file %d lower than for previous frame. Probably filenames are not listed in chronological order of scans.\n\a\n",fr,file);
	    exit(-1);
	  }
	old=frames_midtime_cor[file][fr];
	/*printf("file=%d fr=%d mid_time=%d\n",file,fr,frames_midtime_cor[file][fr]);*/
      }

  return(frames_midtime_cor);

}





/*static float** read_system_response_old(char *fname,int *N_th,int *N_t,double *th_range,double *t_range)
{
  FILE *fp;
  int N,M,i,j;
  int stat;
  float **H;

  fp=fopen(fname,"r");
  if (fp==NULL)
    {
      fprintf(stderr,"Can't open/read system_response file: %s\n\a\n",fname);
      exit(-1);
    }

  fscanf(fp,"%d",N_th);
  fscanf(fp,"%d",N_t);
  fscanf(fp,"%lf",th_range);
  fscanf(fp,"%lf",t_range);
  
  N=*N_th;
  M=*N_t;
  
  H=(float**)multialloc(sizeof(float),2,N,M);

  for (i=0;i<N;i++)
    for (j=0;j<M;j++)
      {
	stat=fscanf(fp,"%f",&H[i][j]);
	if (stat!=1)
	  {
	    fprintf(stderr,"Read error system_response file: stat=%d\n\a\n",stat);
	    exit(-1);
	  }
      }
  fclose(fp);


  return(H);

}
*/

/*static void read_config_file(char *fname, struct sino_info *sino)
{
  FILE *fp;
  int stat;
  char st[255];
  int got_ddet=0,got_rdet=0,got_det_depth=0;
  

  fp=fopen(fname,"r");
  if (fp==NULL)
    {
      fprintf(stderr,"Can't open config_file %s\n\a\n",fname);
      exit(-1);
    }

  while (fscanf(fp,"%s",st)>0)
    {
      if (!strcmp(st,"ddet"))
	{
	  stat=fscanf(fp,"%lf",&sino->ddet);
	  got_ddet=1;
	}
      else if (!strcmp(st,"rdet"))
	{
	  stat=fscanf(fp,"%lf",&sino->rdet);
	  got_rdet=1;
	}
      else if (!strcmp(st,"det_depth"))
	{
	  stat=fscanf(fp,"%lf",&sino->det_depth);
	  got_det_depth=1;
	}
      else
	{
	  fprintf(stderr,"Error reading config-file: Unknown field %s.\n",st);
	  exit(-1);
	}

      if (stat!=1)
	{
	  fprintf(stderr,"Error reading %s value in conf-file. stat=%d.\n\a\n",st,stat);
	  exit(-1);
	}
    }

  if (got_ddet==0)
    {
      fprintf(stderr,"Error reading config-file: No field ddet.\n");
      exit(-1);
    }
  if (got_rdet==0)
    {
      fprintf(stderr,"Error reading config-file: No field rdet.\n");
      exit(-1);
    }
  if (got_det_depth==0)
    {
      fprintf(stderr,"Error reading config-file: No field det_depth.\n");
      exit(-1);
    }


}
*/




float **conv_pix(float **Horg,int N_t,int N_th,double step_t,struct geom_info *geom)
{
  float **H;
  int th,t,i;
  double sum,fact;
  int pix_half_width;
  FILE *fp;

  H=(float**)multialloc(sizeof(float),2,N_th,N_t);
  
  pix_half_width=(int)((geom->delta/step_t - 1)/2 + 0.5);
  if (pix_half_width<0)
    pix_half_width=0;
  
  printf("********* CONVLV: pix_half_width=%d\n\n\n",pix_half_width);

  fact=(double)(2*pix_half_width + 1);

 
  for (th=0;th<N_th;th++)
    {
      for (t=0;t<N_t;t++)
	{
	  sum=0.0;
	  for (i=t-pix_half_width;i<=t+pix_half_width;i++)
	    if ((i>=0) && (i<N_t))
	      sum += Horg[th][i];
	  H[th][t]=(float)(sum/fact);
	}
    }


  fp=fopen("H","w");
  for (th=0;th<N_th;th++)
    {
      for (t=0;t<N_t;t++)
	fprintf(fp,"%1.20f ",H[th][t]);
      fprintf(fp,"\n");
    }
  fclose(fp);

  return(H);

}



void correct_sino_forcbp(struct sino_info *sino,double **Y)
{
  int fr;
  int pr,i,pind,ind,totindex;
  int n_t=112;
  int n_theta=56;
  double cr;

  for (fr=0;fr<sino->no_frames;fr++)
    {
      for(pr=0; pr < n_theta; pr++)
	for(i = 0; i < n_t; i++) /* i=0 is nonsense, same detectors */ 
	  {   
	    pind = pr*n_t;
	    ind = pind+i; /* 1-D index within frame to index into 
			     frame independent correction arrays */
	    totindex=ind+6272*fr; /* index for frame-dependent corrections */

	    if (sino->mask_arr[ind])
	      {
		if (sino->SINO_COR_FLAG)   /* ATTN,DEADTIME,DECAY = frame dependent*/
		  {
		    cr=sino->sino_cor_arr[totindex]; 
		    if (cr<0.05)
		      {
			printf("Warning correct_sino_for_cbp: sino_cor_arr[%d]=%g < 0.05. Setting to 0.05.\n\a\n",totindex,cr);
			cr=0.05;
		      }
		    Y[fr*56+pr][i] /= cr;
		      
		  }


		if (sino->nrm_FLAG) /* detector sensitivity = frame indep*/
		  {
		    cr=sino->nrm_arr[ind];
		    if (cr<0.05)
		      {
			printf("Warning correct_sino_for_cbp: sino->nrm_arr[%d]=%g < 0.05. Setting to 0.05.\n\a\n",ind,cr);
			cr=0.05;
		      }
		    Y[fr*56+pr][i] /= cr;
		      
		  }

	      }
	  }
    }
}



void resort_sino_forcbp(double **Ycbp,struct sino_info *sino,double **Y, int nang)
{
  int nframes;
  double dshift;
  int frameshift,shift,sides_ang,fr,i,j,ind,pr;
  
  int mask[112][56];
  double bigsino111[111][180],sumbigsino111[111][180],dbl_tmp[111][180];
  int bigmask111[111][180],sumbigmask111[111][180],int_tmp[111][180];
  double frsino[112][56];
  int angval;
  int sh_j;
  /*FILE *fp;*/

  nframes=sino->no_frames;
  shift=180/sino->no_frames;
  dshift=180.0/(double)sino->no_frames;
  if (fabs(dshift-(double)shift)>1.0e-8)
    {
      fprintf(stderr,"Error in resort_sino_forcbp: dshift=%f shift=%d\n\a\n",dshift,shift);
      exit(-1);
    }
  printf("cbp sino resort: shift = %d\n",shift);
  sides_ang=nang/2;
  printf("cbp sino resort: nang=%d, using nang/2=%d entries on each side of diamond\n",nang,sides_ang); 


  
  /* make frame mask first of diamond only*/
  for(i=0; i<112; i++)
    for(pr=0; pr<56; pr++)
      {   
	ind = pr*112+i;
	if (sino->mask_arr[ind])
	  mask[i][pr]=1;
	else
	  mask[i][pr]=0;
      }

  /* Now mask out large angles */
  for (i=0;i<56;i++)
    if ( (i<27-sides_ang) || (i>28+sides_ang))
      for (j=0;j<112;j++)
	mask[j][i]=0;
  

  /*fp=fopen("mask","w");
    for (i=0;i<112;i++)
    {
    for (j=0;j<56;j++)
    fprintf(fp,"%d ",mask[i][j]);
    fprintf(fp,"\n");
    }
    fclose(fp);
  */

  /* init summed sinogram and mask to zero */
  for (j=0;j<180;j++)
    for (i=0;i<111;i++)
      {
	sumbigsino111[i][j] = 0.0;
	sumbigmask111[i][j] = 0;
      }

  /* incr. frame moves angle in math neg. direction */
  /* increasing row in diamond also ang in neg dir*/
  /* neg. dir is 89 and upwards in cbp row */ 
  for (fr=0;fr<nframes;fr++)
    {
      frameshift = fr*shift;

      for (i=0;i<56;i++) /* inefficient but safe to take out sgl frame*/
	for (j=0;j<112;j++)
	  {
	    frsino[j][i]=Y[fr*56+i][j]; /* copy frame into new arr*/
	    if (frsino[j][i]<0.0)       /* clip to positive */
	      frsino[j][i]=0.0;         
	    frsino[j][i] = frsino[j][i] * (double)mask[j][i]; /* overlay mask*/
	  }

      for (i=0;i<111;i++)
	for (j=0;j<180;j++)
	  {
	    bigsino111[i][j]=0.0;
	    bigmask111[i][j]=0;
	  }

      angval= (180 - 56)/2; /* taken from Ned's idl code, not checked */

      for (i=0;i<111;i++) /* copy frsino and mask into bigsino and bigmask, cent*/
	for (j=angval;j<angval+56;j++)
	  {
	    bigsino111[i][j] = frsino[i+1][j-angval];
	    bigmask111[i][j] =   mask[i+1][j-angval];
	  }

      /* Now shift bigsino111 and bigmask111 downward in angular
	 direction by frameshift entries,
	 flip displacements of wrapped rows around
      */
 	for (j=0;j<180;j++)
	  {
	    sh_j=j+frameshift;
	    if (sh_j>=180)
	      sh_j -= 180;
	    
	    for (i=0;i<111;i++)
	      {
		dbl_tmp[i][sh_j]=bigsino111[i][j];
		int_tmp[i][sh_j]=bigmask111[i][j];
	      }
	  }
	/* flip first rows until wraparound, copy rest normally*/
	if (frameshift>0)
	  for (j=0;j<frameshift;j++)
	    for (i=0;i<111;i++)
	      {
		bigsino111[i][j] = dbl_tmp[110-i][j]; 
		bigmask111[i][j] = int_tmp[110-i][j]; 
	      }
	/* copy rest normally*/
	if (frameshift<180)
	  for (j=frameshift;j<180;j++)
	    for (i=0;i<111;i++)
	      {
		bigsino111[i][j] = dbl_tmp[i][j]; 
		bigmask111[i][j] = int_tmp[i][j]; 
	      }
	else
	  {
	    fprintf(stderr,"Frameshift= %d >=180??\n\a\n",frameshift);
	    exit(-1);
	  }

	/* FLIP ENTIRE SINOGRAM */
	for (i=0;i<111;i++)
	  for (j=0;j<180;j++)
	    {
	      dbl_tmp[i][j] = bigsino111[110-i][j];
	      int_tmp[i][j] = bigmask111[110-i][j];
	    }
	for (j=0;j<180;j++)
	  for (i=0;i<111;i++)
	    {
	      bigsino111[i][j] = dbl_tmp[i][j];
	      bigmask111[i][j] = int_tmp[i][j];
	    }
	
	/* sum results */
	for (i=0;i<111;i++)
	  for (j=0;j<180;j++)
	    {
	      sumbigsino111[i][j] += bigsino111[i][j];
	      sumbigmask111[i][j] += bigmask111[i][j];
	    }
    } /* end for fr */

  for (i=0;i<111;i++)
    for (j=0;j<180;j++)
      if (sumbigmask111[i][j]==0) 
	{
	  sumbigmask111[i][j]=1000; 
	  if (sumbigsino111[i][j]>1.0e-10)
	    {
	      fprintf(stderr,"Error: summask is zero, but sumbigsino is not\n\a\n");
	      exit(-1);
	    }
	}
  
  for (i=0;i<111;i++)
    for (j=0;j<180;j++)
      Ycbp[j][i] = sumbigsino111[i][j] / (double)sumbigmask111[i][j];
  
  



  

}

  






void read_mat_sino_info(
  MATFile *fin,       /* pointer to file */
  struct sino_info *sino
  )
{
 mxArray *pm;

 pm = matGetArray(fin,"n_theta");
 if(pm != NULL){
  sino->n_theta = (int) mxGetScalar(pm);
  }
 mxDestroyArray(pm);

 pm = matGetArray(fin,"n_t");
 if(pm != NULL){
  sino->n_t = (int) mxGetScalar(pm);
  }
 mxDestroyArray(pm);

 pm = matGetArray(fin,"delta_theta");
 if(pm != NULL){
  sino->delta_theta = mxGetScalar(pm);
  }
 mxDestroyArray(pm);

 pm = matGetArray(fin,"delta_t");
 if(pm != NULL){
  sino->delta_t = mxGetScalar(pm);
  }
 mxDestroyArray(pm);

 pm = matGetArray(fin,"theta_0");
 if(pm != NULL){
  sino->theta_0 = mxGetScalar(pm);
  }
 mxDestroyArray(pm);

 pm = matGetArray(fin,"t_0");
 if(pm != NULL){
  sino->t_0 = mxGetScalar(pm);
  }
 mxDestroyArray(pm);

 sino->dec_a = 1;
 sino->dec_t = 1;


 printf("delta_t = %f\n",sino->delta_t);
 printf("delta_theta = %f\n",sino->delta_theta);
 printf("t_0 = %f\n",sino->t_0);
 printf("theta_0 = %f\n",sino->theta_0);
 printf("n_t = %d\n",sino->n_t);
 printf("n_theta = %d\n",sino->n_theta);

}


